/*
 * milo/identify.c -- Identify the system we're running on
 *
 * Copyright (C) 1994, 1995 by Waldorf Electronics,
 * written by Andreas Busse and Ralf Baechle
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for more
 * details.
 */
#include <linux/string.h>

#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/segment.h>

#include <arc/arc.h>
#include <stand/identify.h>
#include <stand/debug.h>
#include <stand/machspec.h>	/* machines specific routines */

/*
 * Prototypes
 */
int cputype(int id);
char *cpuname(int type);

extern struct bootinfo bi;

unsigned int icache_size     = 0; /* FIXME: cachectl.S needs these */
unsigned int icache_linesize = 0;
unsigned int dcache_size     = 0;
unsigned int dcache_linesize = 0;
unsigned int scache_size     = 0;
unsigned int scache_linesize = 0;



/*
 * Array of bootinfo structures for various machines
 *
 * See linux/include/asm/bootinfo.h for the meaning
 * of each field int the bootinfo struct.
 *
 *
 * IMPORTANT: order is revelant. Each element of this array
 * should correspond to the right MACH_* preprocessor define
 * in linux/include/asm/bootinfo.h If a machine should be
 * added, put an new define in this include file and put a  
 * boot_infotbl[] entry at the position corresponding to value
 * of this define (Is it clear ? :-) 
 *
 * bootinfo struct fields that need to be set (so far):
 * machtype : correspond to a MACH_* define in bootinfo.h
 * memlower
 * memupper
 * scache_size
 * dma_cache_size
 * vram_base
 * tlb_entries
 *
 * nearly all the remaning fields are automagically 
 * updated set identify() (like cache size, cputype)
 *
 */
struct bootinfo boot_infotbl[] ={ 
  /*
   * You don't really hope that Linux boot on such a system, don't you?
   */
   {MACH_UNKNOWN,
    CPU_UNKNOWN,
    KSEG0, 0x80800000,
    0, 0, 0, 0, 0, 0,
    48,				/* true for all R4x00 except R4300i (32) */
    0, 0,
    0, 0,
    0, {{0,}},
    0,
    ""},
   /*
    * RB: This entry is absolutly untested - I don't have
    * even a hint about how the L2 cache works, so I make
    * the guess that it works just like the Tyne board.
    */
   {MACH_DESKSTATION_RPC44, 
    CPU_UNKNOWN, 
    KSEG0, 0x80800000, 
    0, 0, 0, 0, 512*1024, 0, 
    48, 
    512*1024, 0, 
    0, 0, 
    0, {{0,}},
    0xe10a0000, 
    ""},
   {MACH_DESKSTATION_TYNE, 
    CPU_UNKNOWN, 
    KSEG0, 0x80800000, 
    0, 0, 0, 0, 512*1024, 0, 
    48, 
    512*1024, 0, 
    0, 0, 
    0, {{0,}},
    0xe10a0000, 
    ""},
   {MACH_ACER_PICA_61, 
    CPU_UNKNOWN, 
    KSEG0, KSEG0, 
    0, 0, 0, 0, 512*1024, 0, 
    48, 
    0, 0, 
    0, 0, 
    0, {{0,}},
    0xe08b8000,		/* for 80x25, 640x480 */
    ""},
   {MACH_MIPS_MAGNUM_4000,	/* used for Magnum 4000 and Olivetti M700-10 */
    CPU_UNKNOWN, 
    KSEG0, 0x80800000, 
    0, 0, 0, 0, 0, 0, 
    48, 
    0, 0, 
    0, 0, 
    0, {{0,}},
    0xe0800000,		/* FIXME: true for 1024x768 */
    ""}
};


/*
 *  Structure containing infos to get Milo identify
 *  a machine.
 */
struct machine_struct {
  char* SystemId;		/* same as value returned by getSystemId()  */
  
  char* VendorId;		/* VendorId is the same as the value        */
				/* returned by getVendorId() or (char*)NULL */
  
  char* identify_string;	/* string printed by identify() about the machine */
  
  machine_func* machine_ident;	/* specific system identification routine */
				/* (see for Acer Pica 61) */
  
  machine_func* machine_init;	/* specific system initialization before kernel */
				/* launch (see for DeskStation Tyne) */
  
  struct bootinfo* boot_info;	/* pointer on the machine corresponding bootinfo */
				/* structure in boot_infotbl[] */
};


/*
 * Array containing all the machine_struct corresponding to
 * known machines
 *
 * BEWARE!!! Order is revelant. See below in identify() on 
 *           how to add entries 
 */
struct machine_struct machinetbl[] = {
  { "DESKTECH-ARCStation I", (char*)NULL, "DeskStation rPC44", 
    (machine_func*)NULL, (machine_func*)NULL,
    &boot_infotbl[MACH_DESKSTATION_RPC44] },

  { "DESKTECH-TYNE", (char*)NULL, "DeskStation Tyne",
    (machine_func*)NULL, Deskstation_Tyne_init,
    &boot_infotbl[MACH_DESKSTATION_TYNE] },

  { "PICA-61", (char*)NULL, "ACER PICA-61", 
    acer_pica_61_bi, (machine_func*)NULL, 
    &boot_infotbl[MACH_ACER_PICA_61] }, 

  { "Microsoft-Jazz", "OLI00000", "Olivetti M700-10 (Jazz)",
    (machine_func*)NULL, Olivetti_M700_10_init,
    &boot_infotbl[MACH_MIPS_MAGNUM_4000] },

  { "Microsoft-Jazz", (char*)NULL, "Mips Magnum 4000 (Jazz)",
    (machine_func*)NULL, (machine_func*)NULL,
    &boot_infotbl[MACH_MIPS_MAGNUM_4000] },
  /*
   * must be the last element.
   */
  { (char*)NULL, (char*)NULL, " Not identified -- using default parameters", 
    (machine_func*)NULL, (machine_func*)NULL, 
    &boot_infotbl[MACH_UNKNOWN] }
};


extern machine_funcp mach_init;	/* from milo.c */

/*
 * identify()
 */
void identify(void)
{
  char *VendorId;		/* ARC vendor ID */
  char *ProductId;		/* ARC product ID */
  char *SystemId;		/* ARC system ID */
  int  prid;
  int  type;
  int  config;
  int i;
  int machine_idx;		/* index in machinetbl[] corresponding */
				/* of the one identified */

  /*
   * Get system information from ARC BIOS 
   */
  VendorId = getVendorId();
  ProductId = getProductId();
  printf("ARC Vendor  ID          : [");
  for(i=0; i<8; i++)
    printf("%c",(unsigned char)VendorId[i]);
  printf("]\r\n");
  printf("ARC Product ID          : [");
  for(i=0; i<8; i++)
    printf("%02x",(unsigned char)ProductId[i]);
  printf("]\r\n");
  SystemId = getSystemId();
  printf("ARC System  ID          : [%s]\r\n",SystemId);

  /* How machinetbl[] is scanned ? */
  /* We search for a matching SystemId and VendorId fields  */
  /* or for a matching SystemId and a null VendorId         */
  /* So the in machinetbl[] elements like                   */
  /* SystemId = "XXX" VendorId="YYY" *should* come before   */
  /* elements like SystemId = "XXX" VendorId = NULL         */
  /* So specific distinguos can be done and a generic entry */
  /* for each SystemId can be used                          */
  /* See for Olivetti M700 and Magnum 4000                  */                                           
  /* No matching at all permits use of the default entry    */

  for(i=0;i<(sizeof(machinetbl)/sizeof(machinetbl[0]));i++)
    {
      /* BE CAREFULL: the order of this tests is important  */
      /* a NULL pointer shouldn't be passed to strncmp()    */
      /* or be prepared for TLBE. So tests on NULL pointers */
      /* are done before calls to strncmp()                 */
      if ((machinetbl[i].SystemId == (char*)NULL)
	  || ((!strncmp(machinetbl[i].SystemId, SystemId, 8))
	      && ((machinetbl[i].VendorId == (char*)NULL)
		  || (!strncmp(machinetbl[i].VendorId, VendorId, 8))))) 
	{
	  machine_idx = i;
	  break;	      
	}
    }
  memcpy ((void *)(&bi), machinetbl[machine_idx].boot_info, sizeof(bi)); 

  /*
   * Get CPU Info
   */
  printf("ARC CPU ID              : [%s]\r\n",getCpuType());
  prid = read_32bit_cp0_register(CP0_PRID);
  type = cputype(prid);
  printf("Processor type          : [%s I%02x V%d.%d]\r\n",
	 cpuname(type),
	 prid >> 8, (prid & 0xf0) >> 4,prid & 0x0f);

  bi.cputype = type;

  /*
   * Get cache and cache line sizes
   */
  config = read_32bit_cp0_register(CP0_CONFIG);
  icache_size = bi.icache_size = 1 << (12 + ((config & 0xe00) >> 9));
  dcache_size = bi.dcache_size = 1 << (12 + ((config & 0x1c0) >> 6));
  icache_linesize = bi.icache_linesize = 16 << ((config & 0x20) >> 5);
  dcache_linesize = bi.dcache_linesize = 16 << ((config & 0x10) >> 4);
  
  printf("Primary ICache size     : [0x%04x bytes]\r\n",icache_size);
  printf("Primary ICache line size: [0x%04x bytes]\r\n",icache_linesize);
  printf("Primary DCache size     : [0x%04x bytes]\r\n",dcache_size);
  printf("Primary DCache line size: [0x%04x bytes]\r\n",dcache_linesize);

  printf("Secondary cache size    : ");
  if ((bi.scache_size = getSecondaryCacheSize()) != SIZE_UNKNOWN)
    printf("[0x%4x]\r\n",bi.scache_size);
  else
    printf("[unknown]\r\n");

  /* Don't know why, but scache_size in boot_info template seems to be   */
  /* more accurate... ??? (was always overrided in the if {} else if {}  */  
  /* cascade, so... ) */
  bi.scache_size = machinetbl[machine_idx].boot_info->scache_size;

  /* special machine identification stuff ? */
  if (machinetbl[machine_idx].machine_ident != (machine_func*)NULL)
    machinetbl[machine_idx].machine_ident(); 

  mach_init = machinetbl[machine_idx].machine_init;

  printf("ARC system identified as: %s\r\n\r\n", machinetbl[machine_idx].identify_string);
}

/*
 * cputype()
 * Extract cpu type from Processor Id
 * NOTE: This is not complete. Fill in here anything you know about.
 *
 * ID is organized as follows:
 *   Bits 31:16   always 0
 *   Bits 15:08   CPU Implementation
 *   Bits 07:04   CPU Version
 *   Bits 03:00   CPU Revision
 *
 * FIXME: CPU dependent stuff should go into a global array.
 */
int cputype(int id)
{
  switch (id >> 8)
    {
    case 0x01:
      return CPU_R2000;

    case 0x02:
      return CPU_R3000;

    case 0x03:
      return CPU_R6000;
      
    case 0x04:			/* any of R4000xx and R4400xx */
      switch (id & 0xf0)
	{
	case 0x00:		/* R4000 */
	case 0x10:
	case 0x20:
	case 0x30:
	  return CPU_R4000PC;
	  
	case 0x40:		/* R4400 V4.x with daddiu bug */
	case 0x50:		/* R4400 V5.x */
	  return CPU_R4400PC;
	  
	default:
	  return CPU_UNKNOWN;
	}
      break;

    case 0x05:			/* LSI Logic R3000 derivative */
      return CPU_R3000;

    case 0x06:
      return CPU_R6000A;

    case 0x07:
      return CPU_R3051;		/* IDT 3051,3052 etc. */

    case 0x09:
      return CPU_R10000;

    case 0x0a:
      return CPU_R4200;

    case 0x10:
      return CPU_R8000;

    case 0x20:
      return CPU_R4600;		/* R4600 V1.x has an icache bug! */

    case 0x21:
      return CPU_R3000;		/* Sony R3000 derivative */

    case 0x22:
      return CPU_R3000;		/* Toshiba R3000 derivate */

    case 0x23:
      return CPU_R3000;		/* NKK R3000 derivative */
      
    default:
      return CPU_UNKNOWN;
      
    }
}

static char *cpu_name_strings[] = CPU_NAMES;
/*
 * cpuname()
 * Translate cputype into a processor name
 * FIXME: This should also go into a global array of cpu informations.
 *        Some of these CPUs are too stupid (no MMU) to run Linux.
 */
char *cpuname(int type)
{
    if(type > CPU_LAST)
        type = CPU_UNKNOWN;

    return cpu_name_strings[type];
}
